"""
Framework-independent types.
"""


class MLArtifact:
    """A framework-independent machine learning artifact."""


class MLDataset(MLArtifact):
    """A framework-independent machine learning dataset."""


class MLModel(MLArtifact):
    """A framework-independent machine learning model."""
